// -*- C++ -*-

#pragma once

#include "memory.h"
namespace goopax {
namespace impl {
struct resource_impl;


namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_mem_global_barrier(unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_reftest(gpu_writer_base_impl*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_res_ref(resource_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_mem_barrier(resource_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) size_t goopax_4_lds_size(resource_impl*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_uav_create(resource_impl**, unsigned int, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_uav_create_fromtex(resource_impl**, texbuf*, texbuf*, unsigned int, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_lds_create_fromsizescope(resource_impl**, size_t, memory::scope, unsigned int) noexcept;
}
}


inline void local_barrier()
{
    /// Places workgroup-wide barriers on all memory resources. No synchronization takes place between workgroups.
    /// \sa{struct::barrier}
    interface::goopax_4_mem_global_barrier(local_size());
}

inline void global_barrier(unsigned int width_group = num_groups())
{
    /// Places global barriers on all memory resources.
    /// Use with care, this could potentially cause infinite loops if too many workgroups are executed in parallel!
    /// \sa{struct::barrier}
    /// \param width_group If provided, splits the workgroups into blocks of size width_group and only places barriers within the threads in each block. width_group can be any value between 1 and num_groups(). If num_groups() is not divisible by width_group, the last block is smaller.
    interface::goopax_4_mem_global_barrier(width_group * local_size());
}


template<class P>
struct getreftypes;

template<class T>
struct getreftypes<T*>
{
    typedef T  type;
    typedef T& ref;
};
template<class T>
struct getreftypes<const T*>
{
    typedef T        type;
    typedef const T& ref;
};


template<class T, typename CPU_SIZE_TYPE>
struct resource_type_base
{
    using value_type      = typename gettype<T>::gpu;
    using reference       = value_type&;
    using const_reference = const value_type&;
    using baseT           = T;
    static_assert(is_same<CPU_SIZE_TYPE, typename gettype<CPU_SIZE_TYPE>::cpu>::value, "");
    using cpu_size_type  = CPU_SIZE_TYPE;
    using cpu_ssize_type = typename make_signed<cpu_size_type>::type;
    using gpu_size_type  = typename gettype<cpu_size_type>::gpu;
    using gpu_ssize_type = typename gettype<cpu_ssize_type>::gpu;
    using size_type      = gpu_size_type;
    using ssize_type     = gpu_ssize_type;

    unique_ptr<resource_impl, interface::goopax_deleter> impl;
};

template<class T, typename cpu_size_type>
struct resource_with_barrier_type_base : public resource_type_base<T, cpu_size_type>
{
    inline void barrier() const
    {
        interface::goopax_4_mem_barrier(this->impl.get());
    }
};


template<class T, typename CPU_SIZE_TYPE>
struct uav : public resource_with_barrier_type_base<T, CPU_SIZE_TYPE>
{
    static_assert(!is_gpu_type<T>::value, "");
    using BASE = resource_with_barrier_type_base<T, CPU_SIZE_TYPE>;

    using value_type      = typename BASE::value_type;
    using reference       = typename BASE::reference;
    using const_reference = typename BASE::const_reference;
    using baseT           = T;
    using buffer_type     = texture<T, CPU_SIZE_TYPE>;
    using typename BASE::size_type;

    size_type size() const
    {
        size_type ret;
        interface::goopax_4_resource_info(this->impl.get(), ret.impl.get(), 0);
        return ret;
    }

    inline uav(uav&&)   = default;
    inline uav& operator=(uav&&) = default;
    inline uav()
    {
        this->impl.reset(if_call(interface::goopax_4E_uav_create, type_enum<T>::value, gettype<CPU_SIZE_TYPE>::size / gettype<paramtype_t>::size));
    }
    inline uav(texbuf* tex)
    {
        this->impl.reset(if_call(interface::goopax_4E_uav_create_fromtex, tex, (texbuf*)nullptr, type_enum<T>::value, gettype<CPU_SIZE_TYPE>::size / gettype<paramtype_t>::size));
    }
    inline uav(texbuf* tex, texbuf* real)
    {
        this->impl.reset(if_call(interface::goopax_4E_uav_create_fromtex, tex, real, type_enum<T>::value));
    }
    inline ~uav() = default;
};


template<class T, memory::scope SCOPE>
struct lds : public conditional<SCOPE == memory::threadgroup, resource_with_barrier_type_base<T, goopax::impl::Tuint>, resource_type_base<T, goopax::impl::Tuint>>::type
{
    using BASE            = typename conditional<SCOPE == memory::threadgroup, resource_with_barrier_type_base<T, goopax::impl::Tuint>, resource_type_base<T, goopax::impl::Tuint>>::type;
    using value_type      = typename BASE::value_type;
    using reference       = typename BASE::reference;
    using const_reference = typename BASE::const_reference;
    using baseT           = T;
    using typename BASE::cpu_size_type;

    inline cpu_size_type size() const
    {
        return interface::goopax_4_lds_size(this->impl.get());
    }

    inline lds(lds&&)   = default;
    inline lds& operator=(lds&&) = default;
    inline lds(cpu_size_type size0)
    {
        this->impl.reset(if_call(interface::goopax_4E_lds_create_fromsizescope, (size_t)size0, SCOPE, type_enum<T>::value));
    }
    inline ~lds() = default; // FIXME: Barrier?
};


}
using impl::global_barrier;
using impl::local_barrier;

}
