// -*- C++ -*-

#pragma once

#include "defs.h"
#include "stdinc.h"


/*
namespace goopax
{
  namespace impl
  {
    template <typename T, typename P> struct resource_ref;
  }
}
*/


namespace goopax {
namespace impl {


template<size_t n, class T>
struct isplain<std::array<T, n>> : public isplain<T>
{};
#if GOOPAX_HAVE_DEBUG_STL
template<size_t n, class T>
struct isplain<std::__debug::array<T, n>> : public isplain<T>
{};
#endif


template<typename... T>
struct testhelper
{
    using type = void;
};


template<typename T, typename = void, typename = void, typename = void>
struct is_derived_from_vector : false_type
{};

template<typename T>
struct is_derived_from_vector<T,
                              typename testhelper<typename T::value_type>::type,
                              typename testhelper<typename T::allocator_type>::type> :

    public integral_constant<bool, is_base_of<std::vector<typename T::value_type, typename T::allocator_type>, T>::value
#if GOOPAX_HAVE_DEBUG_STL
                                       || is_base_of<std::__debug::vector<typename T::value_type, typename T::allocator_type>, T>::value
#endif
                             >

{};


template<typename T, typename = void, typename = void>
struct is_derived_from_array2 : false_type
{};

template<typename T>
struct is_derived_from_array2<T,
                              typename testhelper<typename T::value_type>::type,
                              typename testhelper<typename T::value_type[5]>::type> : public integral_constant<bool, std::is_base_of<std::array<typename T::value_type, sizeof(T) / sizeof(typename T::value_type)>, T>::value
#if GOOPAX_HAVE_DEBUG_STL
                                                                                                                         || std::is_base_of<std::__debug::array<typename T::value_type, sizeof(T) / sizeof(typename T::value_type)>, T>::value
#endif
                                                                                                               >
{};

template<typename T>
struct is_derived_from_array : is_derived_from_array2<T>
{
};

template<typename T>
struct is_derived_from_array<std::array<T, 0>> : public true_type
{};
#if GOOPAX_HAVE_DEBUG_STL
template<typename T>
struct is_derived_from_array<std::__debug::array<T, 0>> : public true_type
{};
#endif


template<typename T>
struct is_arithmetic_container : integral_constant<bool, (is_derived_from_vector<T>::value || is_derived_from_array<T>::value)>
{};


template<typename T>
struct is_string : false_type
{};
template<typename... T>
struct is_string<std::basic_string<T...>> : true_type
{};
#if GOOPAX_HAVE_DEBUG_STL
template<typename C, typename T, typename A>
struct is_string<__gnu_debug::basic_string<C, T, A>> : true_type
{};
#endif

template<typename T>
struct is_container : public integral_constant<bool, (is_arithmetic_container<T>::value || is_string<T>::value)>
{};


template<class V, class B>
struct changetype;

template<typename A, class B>
struct changetype<std::vector<A>, B>
{
    using type = std::vector<B>;
};
#if GOOPAX_HAVE_DEBUG_STL
template<typename A, typename ALLOC, class B>
struct changetype<std::__debug::vector<A, ALLOC>, B>
{
    using type = std::__debug::vector<B, typename ALLOC::template rebind<B>::other>;
};
#endif
template<class A, size_t N, class B>
struct changetype<std::array<A, N>, B>
{
    using type = std::array<B, N>;
};
#if GOOPAX_HAVE_DEBUG_STL
template<class A, size_t N, class B>
struct changetype<std::__debug::array<A, N>, B>
{
    using type = std::__debug::array<B, N>;
};
#endif
template<typename A, typename B>
struct changetype<std::basic_string<A>, B>
{
    using type = std::basic_string<B>;
};
#if GOOPAX_HAVE_DEBUG_STL
template<typename A, typename ALLOC, typename B>
struct changetype<__gnu_debug::basic_string<A, std::char_traits<A>, ALLOC>, B>
{
    using type = __gnu_debug::basic_string<B, std::char_traits<B>, typename ALLOC::template rebind<B>::other>;
};
#endif


template<typename T>
integral_constant<size_t, ~(size_t(0))> fixed_size_func(const T&);
template<typename T, size_t N>
integral_constant<size_t, N> fixed_size_func(const std::array<T, N>&);
#if GOOPAX_HAVE_DEBUG_STL
template<typename T, size_t N>
integral_constant<size_t, N> fixed_size_func(const std::__debug::array<T, N>&);
#endif

template<typename T>
struct get_fixed_size : public decltype(fixed_size_func(declval<T>()))
{};


template<typename A, typename B>
inline void estest(const A& a, const B& b)
{
    static_assert(get_fixed_size<A>::value == ~(size_t(0)) || get_fixed_size<B>::value == ~(size_t(0)) || get_fixed_size<A>::value == get_fixed_size<B>::value,
                  "Array size does not match!");

    if (a.size() != b.size())
    {
        // Provoking failure for debug containers. Will be optimized away for non-debug containers.
        (void)*a.end();
        (void)*b.end();
    }
}


template<typename FROM, typename TO>
void convert_assign(TO& to, const FROM& from);

template<typename FROM, typename TO>
void convert_assign_help(TO& to, const FROM& from, false_type, false_type)
{
    to = static_cast<TO>(from);
}

template<typename FROM, typename TO>
void convert_assign_help(TO& to, const FROM& from, false_type, true_type)
{
    to.resize(from.size());

    for (size_t k = 0; k < to.size(); ++k)
    {
        convert_assign(to[k], from[k]);
    }
}

template<typename FROM, typename TO>
void convert_assign_help(TO& to, const FROM& from, true_type, true_type)
{
    estest(to, from);

    for (size_t k = 0; k < to.size(); ++k)
    {
        convert_assign(to[k], from[k]);
    }
}

template<typename FROM, typename TO>
void convert_assign(TO& to, const FROM& from)
{
    convert_assign_help(to, from, is_derived_from_array<TO>(), is_container<TO>());
}

template<typename TO, typename FROM>
TO convert(const FROM& from)
{
    TO to;
    convert_assign(to, from);
    return to;
}


}
}
