#pragma once

#include "buffer.h"


namespace goopax {
namespace impl {
namespace interface {
#if GOOPAX_HAVE_METAL && defined(__OBJC__)
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_get_metal_texbuf(void** ret, texbuf* buf) noexcept;
#endif
}

#if GOOPAX_HAVE_METAL && defined(__OBJC__)
template<typename T, typename SIZE_TYPE>
id<MTLBuffer> get_metal_mem(const buffer<T, SIZE_TYPE>& buf)
{
    /// \ingroup Metal
    /// Returns Metal memory handle for the goopax buffer
    /// \param buf buffer
    return id<MTLBuffer>(if_call(interface::goopax_4E_get_metal_texbuf, buf.get_buf()));
}

template<unsigned int DIM, typename T, bool normalized>
id<MTLTexture> get_metal_mem(const image_buffer<DIM, T, normalized>& image)
{
    /// \ingroup Metal
    /// Returns Metal memory handle for the goopax image
    /// \param buf image buffer
    return id<MTLTexture>(if_call(interface::goopax_4E_get_metal_texbuf, image.get_buf()));
}

template<unsigned int DIM, typename T, bool normalized>
id<MTLTexture> get_metal_mem(const image_array_buffer<DIM, T, normalized>& image_array)
{
    /// \ingroup Metal
    /// Returns Metal memory handle for the goopax image array
    /// \param buf image array buffer
    return id<MTLTexture>(if_call(interface::goopax_4E_get_metal_texbuf, image_array.get_buf()));
}
#endif
}

#if GOOPAX_HAVE_METAL && defined(__OBJC__)
using impl::get_metal_mem;
#endif
}
